#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <stdio.h>
#include <pthread.h>
#include <string.h>
#include <poll.h>

#include "Connexion_client_tcp.h"
#include "Communication_tcp.h"
#include "Gestion_requete.h"
#include "Hole_punching.h"

/*Connexion au serveur TCP*/
int connexion_serveur(int port, char* hote)
{
#ifdef DEBUG 
    printf("[Client] [connexion_client_tcp.c] [connexion_serveur] Début de la connexion au serveur d'adresse : '%s' sur le port '%d'\n",hote,port);
#endif
    
    int s;
    struct sockaddr_in adresse;
    socklen_t taille=sizeof adresse;
    
    /* Creation d'une socket */
    s=socket(PF_INET,SOCK_STREAM,0);
    if(s<0){
        perror("connexionServeur.socket");
        exit(-1);
    }
    
#ifdef DEBUG 
    printf("[Client] [connexion_client_tcp.c] [connexion_serveur] Socket créée avec le numéro : '%d'\n",s);
#endif
    
    /*Initialisation des options nécessaires pour le hole_punching*/
    int vrai =1;
    if(setsockopt(s,SOL_SOCKET,SO_REUSEADDR,&vrai,sizeof(vrai))<0){
        perror("initialisationServeur.setsockopt (REUSEADDR)");
        exit(-1);
    }
    if(setsockopt(s,SOL_SOCKET,SO_REUSEPORT,&vrai,sizeof(vrai))<0){
        perror("initialisationServeur.setsockopt (REUSEPORT)");
        exit(-1);
    }
    /*Récupération des informations serveur */
    if(nomVersAdresse(hote,(void *)&adresse)<0) return -1; 
    adresse.sin_port=htons(port);
        
#ifdef DEBUG 
    printf("[Client] [connexion_client_tcp.c] [connexion_serveur] addr créée avec les attributs suivants : IP : '%s' et port : '%d'\n",inet_ntoa(adresse.sin_addr), ntohs(adresse.sin_port));
#endif
    
    /* Connexion de la socket à l'hôte */
    if(connect(s,(struct sockaddr *)&adresse,taille)<0) return -1;
    else return s;
}


/*Récupère les informations nécessaires pour la connexion au serveur et les rentre dans une struct sockaddr prête à l'emploi, il suffit de lui donner l'addresse IP du serveur au format char* */
int nomVersAdresse(char *hote,struct sockaddr_storage *padresse)
{
    struct addrinfo *resultat,*origine;
    int statut=getaddrinfo(hote,NULL,NULL,&origine);
    if(statut==EAI_NONAME) return -1;
    if(statut<0){ perror("nomVersAdresse.getaddrinfo"); exit(EXIT_FAILURE); }
    struct addrinfo *p;
    for(p=origine,resultat=origine;p!=NULL;p=p->ai_next){
        if(p->ai_family==AF_INET6){ resultat=p; break; }
        memcpy(padresse,resultat->ai_addr,resultat->ai_addrlen);
    }
    return 0;
}


/*Fonction principale de communication avec le serveur TCP*/
void ClientTCP(char* adresse, short int service, int s){
    printf("[client] [connexion_client_tcp.c] [ClientTCP] Lancement serveur TCP addresse: '%s' sur le port : '%d'\n", adresse, service);
    /* Connexion au serveur TCP rendez-vous*/
    s = connexion_serveur(service,adresse);
    
#ifdef DEBUG
    printf("[client] [Connexion_client_tcp] [ClientTCP] Connecté sur la socket '%d'\n",s);
#endif
    
    if(s<0){ 
        fprintf(stderr,"Erreur de connexion au serveur\n");
        perror("Client.connexion_client_tcp.ClientTCP.connect");
        exit(EXIT_FAILURE); 
    }    
    
    affichage_menu();
    
    char* pseudo = (char *) malloc(sizeof(char )*TAILLE_PSEUDO);
    /* Boucle de communication avec le serveur */
    struct pollfd descripteurs[2];
    descripteurs[0].fd=s;       //fd au serveur
    descripteurs[0].events=POLLIN;
    descripteurs[1].fd=0;       //fd=0 signifie provenant du même terminal
    descripteurs[1].events=POLLIN;
    Identification_serveur(s,pseudo); //fonction permettant l'identification du client auprès du serveur
    
    /*Permet l'envoi et la réception des messages avec le serveur le plus facilement et le plus rapidement possible */
    int Client_on = 1;
    while(Client_on){
        int nb=poll(descripteurs,2,-1);
        char* tampon = (char *) malloc(sizeof(char )*MAX_TAMPON);
        if(nb<0){ perror("main.poll"); exit(EXIT_FAILURE); }
        
        /* Regarde si le serveur a envoyé un message pour l'afficher sur le terminal à travers la variable tampon */
        if((descripteurs[0].revents&POLLIN)!=0){
            int taille=read(s,tampon,MAX_TAMPON);
            if(taille<=0) {
                printf("[client] [connexion_client_tcp.c] [ClientTCP] Problème réception du message par le serveur\n");
                Client_on = 0;
                break;
            }
            
#ifdef DEBUG
            printf("[client] [connexion_client_tcp.c] [ClientTCP] Message reçu du serveur avant gestion requete serveur : '%s'\n",tampon);
#endif
            /*Verifie que le message ne soit pas une commande en provenance du serveur*/
            int requete_serveur = gestion_requete_serveur(tampon);
            if( requete_serveur == 0){
                printf("[client] [connexion_client_tcp.c] [ClientTCP] Réception d'un message du serveur :");
                write(1,tampon,taille);
            }
            else if(requete_serveur == 2){
                hole_punching(tampon,s);
                affichage_menu();
            }
            else{
                Client_on = 0;
            }
        }
        
        /* Regarde si un message est rentré dans la console pour pouvoir l'envoyer au serveur */
        if((descripteurs[1].revents&POLLIN)!=0){
            int taille=read(0,tampon,MAX_TAMPON);
            if(taille<=0){
                printf("[client] [connexion_client_tcp.c] [ClientTCP] Problème d'envoi du message au serveur\n");
                Client_on = 0;
                break;
            }
            tampon[taille] = '\0';
            /*traduit le message entré en commande pour le serveur*/
            int requete_client = gestion_requete_client(tampon,pseudo);
            taille = strlen(tampon);
            if(requete_client == -1){
                write(s,tampon,taille);
                Client_on = 0;
            }
            else if(requete_client == 1){
                write(s,tampon,taille);
            }
                      
#ifdef DEBUG
                printf("[client] [connexion_client_tcp.c] [ClientTCP] Message envoyé au serveur: '%s'\n",tampon);
#endif
        
        }
        free(tampon);
    }
    
    /* On termine la connexion */
    free(pseudo);
    shutdown(s,SHUT_RDWR);
    close(s);
    printf("\nArret du client \n");
}
