#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <stdio.h>
#include <pthread.h>
#include <string.h>
#include <poll.h>
#include <arpa/inet.h>     

#include "Connexion_client_udp_6.h"
#include "Communication_udp_6.h"
#include "Gestion_requete_6.h"

static int Hole_punching_on;

int connexion_serveur_udp(char *hote,char *service){
    
#ifdef DEBUG	
    printf("[connexion_client_udp_6.c] [connexion_serveur_udp] Début avec hote : '%s' et service : '%s'\n", hote, service);
#endif
    
    struct addrinfo precisions,*resultat,*origine;
    int statut;
    int s;
    
    /* Creation de l'adresse de socket */
    memset(&precisions,0,sizeof precisions);
    precisions.ai_family=PF_INET6;
    precisions.ai_socktype=SOCK_DGRAM;
    statut=getaddrinfo(hote,service,&precisions,&origine);
    if(statut!=0){ 
        perror("Connexion_client_udpgenerique.getaddrinfo"); 
        return -1;
        exit(EXIT_FAILURE);
    }
    struct addrinfo *p;
    for(p=origine,resultat=origine;p!=NULL;p=p->ai_next){        
        /* Creation d'une socket */
        s=socket(resultat->ai_family,resultat->ai_socktype,resultat->ai_protocol);
    }
    if(s<0){ perror("Connexion_client_udpgenerique.socket"); exit(EXIT_FAILURE); }
    
    /* Option sur la socket */
    int vrai=1;
    if(setsockopt(s,SOL_SOCKET,SO_KEEPALIVE,&vrai,sizeof(vrai))<0){
        perror("initialisationServeurUDPgenerique.setsockopt (BROADCAST)");
        exit(-1);
    }
    if(setsockopt(s,SOL_SOCKET,SO_REUSEADDR,&vrai,sizeof(vrai))<0){
        perror("initialisationServeurUDPgenerique.setsockopt (REUSEADDR)");
        exit(-1);
    }
    if(setsockopt(s,SOL_SOCKET,SO_REUSEPORT,&vrai,sizeof(vrai))<0){
        perror("initialisationServeur.setsockopt (REUSEPORT)");
        exit(-1);
    }
#ifdef DEBUG	
    printf("[connexion_client_udp_6.c] [connexion_serveur_udp] fin\n");
#endif
    /* Liberation de la structure d'informations */
    freeaddrinfo(origine);
    
    return s;
}

void ClientUDP(char* adresse, short int service, int s){
    printf("[client] [Connexion_client_udp_6.c] [ClientUDP] Lancement serveur UDP à l'adresse : '%s' sur le port : '%hd'\n", adresse, service);
    struct sockaddr_in6 addr;
    socklen_t taille=sizeof(addr);
    char* char_service = (char *) malloc(sizeof(char )*5);
    char* pseudo = (char *) malloc(sizeof(char )*TAILLE_PSEUDO);
    
    addr.sin6_family = AF_INET6; 
    addr.sin6_port = htons(service);
    addr.sin6_flowinfo=0;
    addr.sin6_scope_id=0; 
    inet_pton(AF_INET6, adresse, &addr.sin6_addr);
    char *str=malloc(sizeof(char)*INET6_ADDRSTRLEN);
    memset(str,0,sizeof(char)*INET6_ADDRSTRLEN);
    inet_ntop(AF_INET6, &(addr.sin6_addr), str, sizeof(char)*INET6_ADDRSTRLEN);
#ifdef DEBUG	
    printf("[client] [Connexion_client_udp_6.c] [ClientUDP] info_adresse : port : %d , addresse : %s\n",ntohs(addr.sin6_port),str);
#endif
    free(str);
    sprintf(char_service,"%d",service);
    
#ifdef DEBUG	
    printf("[client] [Connexion_client_udp_6.c] [ClientUDP] info_adresse : char_service : %s , adresse : %s\n",char_service,adresse);
#endif
    
    /* Connexion au serveur */
    s = connexion_serveur_udp(adresse,char_service);
    
    free(char_service);
    
    /* Boucle de communication avec le serveur */
    struct pollfd descripteurs[2];
    descripteurs[0].fd=s;       //fd au serveur
    descripteurs[0].events=POLLIN;
    descripteurs[1].fd=0;       //fd=0 signifie provenant du même terminal
    descripteurs[1].events=POLLIN;
    Identification_serveur_udp(s, addr, pseudo); //fonction permettant l'identification du client auprès du serveur
    affichage_menu();
    /*Permet l'envoi et la réception des messages avec le serveur le plus facilement et le plus rapidement possible */
    int Client_on = 1;
    while(Client_on){
        poll(descripteurs,2,-1);
        char* tampon = (char *) malloc(sizeof(char )*MAX_TAMPON);
        /* Regarde si le serveur a envoyé un message pour l'afficher sur le terminal à travers la variable tampon */
        if((descripteurs[0].revents&POLLIN)!=0){
            int rcv_data = recvfrom(s,tampon,MAX_TAMPON,0,(struct sockaddr *)&addr,&taille);
            if(rcv_data<=0) {
                printf("[client] [connexion_client_udp_6.c] [ClientUDP] Problème réception du message par le serveur\n");
                break;
            }
                                
#ifdef DEBUG
        printf("[client] [Connexion_client_udp_6.c] [Client UDP] Message recu avant gestion_requete_serveur '%s'\n", tampon);
#endif
       
            int requete_serveur = gestion_requete_serveur(tampon);
            if( requete_serveur == 0){
                printf("[client] [connexion_client_udp_6.c] [ClientUDP] Réception d'un message du serveur :");
                write(0, tampon, rcv_data);
            }
            if( requete_serveur == 2){
                hole_punching_udp(tampon,s);
                affichage_menu();
            }
            if( requete_serveur == 1){
                Client_on = 0;
            }
        }
        
        /* Regarde si un message est rentré dans la console pour pouvoir l'envoyer au serveur */
        if((descripteurs[1].revents&POLLIN)!=0){
            int taille=read(0,tampon,MAX_TAMPON);
            if(taille<=0){
                printf("[client] [connexion_client_udp_6.c] [ClientUDP] Problème d'envoi du message au serveur\n");
                break;
            }
            tampon[taille] = '\0';
            int requete_client = gestion_requete_client(tampon,pseudo);
            taille = strlen(tampon);
            if(requete_client == -1){
                envoie_message_udp(tampon, s, taille,addr);
                
#ifdef DEBUG
                printf("[client] [connexion_client_udp_6.c] [ClientUDP] Message envoyé au serveur : '%s'\n",tampon);
#endif
                
                Client_on = 0;
            }
            else if(requete_client == 1){
                envoie_message_udp(tampon, s, taille,addr);
                
#ifdef DEBUG
                printf("[client] [connexion_client_udp_6.c] [ClientUDP] Message envoyé au serveur : '%s'\n",tampon);
#endif
                
            }
        }
        free(tampon);
    }
    
    /* Fermeture de la socket d'envoi */
    close(s);
    free(pseudo);
    printf("\nArret du client \n");
}

void hole_punching_udp(char * message, int socket){
    Hole_punching_on = 1;
    
#ifdef DEBUG
    printf("[client] [Connexion_client_udp_6.c] [hole_punching_udp] Demande de hole punching du serveur avec le message : '%s'\n",message);
#endif
    
    char *chaine = NULL;
    chaine = strchr(message,' ');
    char* client_B = malloc(sizeof(char)*TAILLE_PSEUDO);
    int i = 1;
    while(chaine[i] != '\n'){   //Récupération du login que le serveur a envoyé
        client_B[i-1] = chaine[i];
        i++;
    }
    

        
    client_B[i-1] = '\0';
    chaine[0] = 'a';

#ifdef DEBUG
        printf("[client] [Connexion_client_udp_6.c] [hole_punching_udp] Login du peer avec qui nous allons initier la connexion '%s'\n Message restant à traiter chaine : '%s' et message : '%s'\n",client_B,chaine, message);
#endif

    chaine = strchr(message,' ');
    char* IP_client_B = malloc(sizeof(char)*25);
    i = 1;
    while(chaine[i] != '\n'){   //Récupération de l'IP que le serveur a envoyé
        IP_client_B[i-1] = chaine[i];
        i++;
    }
    IP_client_B[i-1] = '\0';
    chaine[0] = 'a';

#ifdef DEBUG
        printf("[client] [Connexion_client_udp_6.c] [hole_punching_udp] IP du peer avec qui nous allons initier la connexion '%s'\n Message restant à traiter chaine : '%s' et message : '%s'\n",IP_client_B,chaine, message);
#endif

    
    
    chaine = strchr(message,' ');
    char* port_client_B = malloc(sizeof(char)*5);
    i = 1;
    int port = 0;
    while(chaine[i] != '\n'){   //Récupération du port que le serveur a envoyé
        port_client_B[i-1] = chaine[i];
        i++;
    }
    port_client_B[i-1] = '\0';

#ifdef DEBUG
        printf("[client] [Connexion_client_udp_6.c] [hole_punching_udp] Port du peer avec qui nous allons initier la connexion '%s'\n Message restant à traiter chaine : '%s' et message : '%s'\n",port_client_B,chaine, message);
#endif

    i = 0;
    while(port_client_B[i] >=48 && port_client_B[i] <=57 ){
        port =port*10 +( (short int ) port_client_B[i]-48);
        i++;
    }

#ifdef DEBUG
        printf("[client] [Connexion_client_udp_6.c] [hole_punching_udp] Port client B en int : '%d'\n",port);
#endif

    
    struct sockaddr_in6 addr_client_B;
    
    addr_client_B.sin6_family = AF_INET6; 
    addr_client_B.sin6_port = htons(port);
    inet_pton(AF_INET6,IP_client_B, &addr_client_B.sin6_addr);
    char *str=malloc(sizeof(char)*INET6_ADDRSTRLEN);
    memset(str,0,sizeof(char)*INET6_ADDRSTRLEN);
    inet_ntop(AF_INET6, &(addr_client_B.sin6_addr), str, sizeof(char)*INET6_ADDRSTRLEN);
#ifdef DEBUG
        printf("[client] [Connexion_client_udp_6.c] [hole_punching_udp] Information du peer avec qui nous allons initier la connexion, son IP : '%s' et son port : '%d'\n" ,str,ntohs(addr_client_B.sin6_port));
#endif
    free(str);
 /*   int statut = connexion_hole_punching(addr_client_B,socket);
    if(statut == -1){
        printf("Connexion à %s a échoué\n",client_B);
        return;
    }
    pthread_t pid;
    struct information_hole_punching *peerB = malloc(sizeof(struct information_hole_punching));
    peerB->addr = addr_client_B;
    peerB->socket = socket;
     statut = pthread_create(&pid,NULL,keep_alive,(void*) peerB); Désactive les keep alive*/
    communication_peer_to_peer(client_B,addr_client_B,socket);
}    
    
int connexion_hole_punching( struct sockaddr_in6 addr, int socket){
    char msg[100] = "Connexion!!!\n";
    char *str=malloc(sizeof(char)*INET6_ADDRSTRLEN);
    memset(str,0,sizeof(char)*INET6_ADDRSTRLEN);
    inet_ntop(AF_INET6, &(addr.sin6_addr), str, sizeof(char)*INET6_ADDRSTRLEN);
#ifdef DEBUG
        printf("[client] [Connexion_client_udp_6.c] [connexion_hole_punching] Tentative de connexion au peer avec le message : '%s' son IP : '%s' et son port : '%d'\n",msg,str,ntohs(addr.sin6_port) );
#endif

    int i = 0;
    while(1){
        char *buffer = malloc(sizeof(char)*100);
        envoie_message_udp(msg,socket,100,addr);
        i = reception_message_udp(buffer,socket);
        free(buffer);
        if(i == 1){
                
#ifdef DEBUG
        printf("[client] [Connexion_client_udp_6.c] [connexion_hole_punching] Connexion avec le peer établie\n");
#endif
        
            return 1;
        }
#ifdef DEBUG
        static int num = 1;
        printf("[client] [Connexion_client_udp_6.c] [connexion_hole_punching] Tentative numéro '%d' de connexion au peer avec le message : '%s' son IP : '%s' et son port : '%d'\n",num++, msg,str,ntohs(addr.sin6_port) );
#endif
        sleep(1);
    }
    free(str);
    return -1;
}

void* keep_alive(void* arg){
    struct information_hole_punching *clientB = (struct information_hole_punching*) arg;
    char msg[15] = "KeepAlive!!!\n";
    char *str=malloc(sizeof(char)*INET6_ADDRSTRLEN);
    memset(str,0,sizeof(char)*INET6_ADDRSTRLEN);
    inet_ntop(AF_INET6, &(clientB->addr.sin6_addr), str, sizeof(char)*INET6_ADDRSTRLEN);
#ifdef DEBUG
        printf("[client] [Connexion_client_udp_6.c] [keep_alive] Début du keep alive avec le correspondant possédant les informations  suivantes : socket : '%d' IP: '%s' et port : '%d'\n",clientB->socket ,str,clientB->addr.sin6_port );
#endif
        
    while(Hole_punching_on == 1){
        envoie_message_udp(msg,clientB->socket,15,clientB->addr);
        sleep(10);
    }
    free(arg);
    free(str);
    pthread_exit(0);
}

void communication_peer_to_peer(char* client_B, struct sockaddr_in6 addr, int socket){   
    char *str=malloc(sizeof(char)*INET6_ADDRSTRLEN);
    memset(str,0,sizeof(char)*INET6_ADDRSTRLEN);
    inet_ntop(AF_INET6, &(addr.sin6_addr), str, sizeof(char)*INET6_ADDRSTRLEN);
#ifdef DEBUG
        printf("[client] [Connexion_client_udp_6.c] [communication_peer_to_peer] Début de la communication peer to peer avec le correspondant possédant les informations suivantes : socket : '%d' IP: '%s' et port : '%d'\n", socket,str, ntohs(addr.sin6_port));
#endif
    struct pollfd descripteurs_Peer_To_Peer[2];
    descripteurs_Peer_To_Peer[0].fd=socket;       //fd du client
    descripteurs_Peer_To_Peer[0].events=POLLIN;
    descripteurs_Peer_To_Peer[1].fd=0;       //fd=0 signifie provenant du même terminal
    descripteurs_Peer_To_Peer[1].events=POLLIN;
    socklen_t taille = sizeof(addr);
    /*Permet l'envoi et la réception des messages avec le serveur le plus facilement et le plus rapidement possible */
    while(Hole_punching_on){
        poll(descripteurs_Peer_To_Peer,2,-1);
        char* tampon = (char *) malloc(sizeof(char )*MAX_TAMPON);
        /* Regarde si le peer a envoyé un message pour l'afficher sur le terminal à travers la variable tampon */
        if((descripteurs_Peer_To_Peer[0].revents&POLLIN)!=0){
            int rcv_data = recvfrom(socket,tampon,MAX_TAMPON,0,(struct sockaddr *)&addr,&taille);
            if(rcv_data<=0) {
                printf("[client] [Connexion_client_udp_6.c] [communication_peer_to_peer] Problème de réception du message du peer\n");
                Hole_punching_on = 0;
                free(tampon);
                break;
            }
            tampon[rcv_data] = '\0';
            int message_peerB = gestion_message_peerB_recu(tampon);
            if( message_peerB == 1){
                int taille_pseudo_client = strlen(client_B);
                char* message_a_afficher = malloc(sizeof(char)*(rcv_data + taille_pseudo_client + 5));
                int n = sprintf(message_a_afficher,"%s : %s", client_B, tampon);
                if(n < 0){
                    printf("[client] [Connexion_client_udp_6.c] [communication_peer_to_peer] Problème du sprintf pour afficher le message sur le terminal\n");
                    Hole_punching_on = 0;
                    free(tampon);
                    break;
                }
                message_a_afficher[n] = '\0';
                write(0, message_a_afficher, strlen(message_a_afficher));
                free(message_a_afficher);
            }
            if( message_peerB == -1){
                printf("[client] [Connexion_client_udp_6.c] [communication_peer_to_peer] Déconnexion de la communication peer to peer et retour à la communication avec le serveur\n");
                Hole_punching_on = 0;
            }
            if(message_peerB == 0){
#ifdef DEBUG
                printf("[client] [Connexion_client_udp_6.c] [communication_peer_to_peer] Réception du keep_alive du client à addr : '%s' au port '%d'\n",str, ntohs(addr.sin6_port));
#endif
            }
        }
        
        /* Regarde si un message est rentré dans la console pour pouvoir l'envoyer au Peer */
        if((descripteurs_Peer_To_Peer[1].revents&POLLIN)!=0){
            int snd_data=read(0,tampon,MAX_TAMPON);
            if(snd_data<=0){
                printf("[client] [Connexion_client_udp_6.c] [communication_peer_to_peer] Problème d'envoi de message au peer\n");
                Hole_punching_on = 0;
                free(tampon);
                break;
            }
            tampon[snd_data] = '\0';
            int message_pour_peerB = gestion_message_peerB_envoie(tampon);
            //snd_data = strlen(tampon);
            if(message_pour_peerB == -1){
                envoie_message_udp(tampon, socket, snd_data,addr);
                printf("[client] [Connexion_client_udp_6.c] [communication_peer_to_peer] Déconnexion de la communication peer to peer\n");
                Hole_punching_on = 0;
            }
            else if(message_pour_peerB == 1){
                int statut = envoie_message_udp(tampon, socket, snd_data,addr);
                if(statut == -1){
                    printf("[client] [Connexion_client_udp_6.c] [communication_peer_to_peer] Perte de connexion avec notre peer\n");
                    Hole_punching_on = 0;
                }
                else{
#ifdef DEBUG
                    printf("[client] [Connexion_client_udp_6.c] [communication_peer_to_peer] Message envoyé au peer '%s' : '%s'\n",client_B, tampon);
#endif
                }
            }
        }
        free(tampon);
    }
    free(str);
}

int gestion_message_peerB_recu(char* msg){
    char *i = strstr(msg, "exit\n");
    if(i != NULL){
        return -1;
    }
    i = strstr(msg,"KeepAlive!!!\n");
    if(i != NULL){
        return 0;
    }
    return 1;
}

int gestion_message_peerB_envoie(char* msg){
    char* i = strstr(msg, "exit\n");
    if(i != NULL){
        return -1;
    }
    return 1;
}
