

#include "Gestion_Client.h"

/*Initialisation afin d'éviter tous problèmes avec notre liste de clients connectés */
void initialisation_liste_client(){
    liste_client[MAX_CLIENT_CONNECTE].socket = -1;
    for(int j =0; j < TAILLE_PSEUDO ; j++){
        liste_client[MAX_CLIENT_CONNECTE].pseudo[j] = 0;
    }
}

/*Fonction de traitement des clients TCP*/
void *gestionClient(void *arg){
    
#ifdef DEBUG
    printf("[serveur.c] [gestionClient] Arrivé d'un client\n");
#endif
    
    info_client* s =(info_client*) arg;
    int socket = s->socket;

    /* Obtient une structure de fichiers */
    
    char ligne[MAX_LIGNE] = "";
    int j = 0;
    while((j=read(socket,ligne,MAX_LIGNE))>0){
        ligne[j] = '\0';                             
#ifdef DEBUG
        printf("[Serveur] [Gestion_Client.c] [gestionClient] Message reçu avant gestion_requete '%s'\n", ligne);
#endif
       
        int statut = gestion_requete(ligne,s,"TCP");
        if(statut == 0){
            printf("%s : %s",s->pseudo,ligne);
        }
        if(statut == -1){
            break;
        }
        if(statut == -2 ){
            printf("Message indésirable reçu : '%s'\n",ligne);
            char* message_erreur = malloc(sizeof(char)*MAX_LIGNE);
            strcpy(message_erreur,"Veuillez utiliser notre Client pour communiquer avec notre serveur\n");
            write(socket,message_erreur,MAX_LIGNE);
                                     
#ifdef DEBUG
            printf("[Serveur] [Gestion_Client.c] [gestionClient] Message d'erreur envoyé: '%s'\n", message_erreur);
#endif
        
            break;
        }
        for(int i = 0 ; i< MAX_LIGNE ; i++){
            ligne[i] = 0;
        }
    }
    /* Termine la connexion avec libération de mémoire, fermeture du socket, libération de la place prise dans le répertoire liste_client et fermeture du thread de traitement */
    Deconnexion_client(*s);
    close(socket);
    free(arg);
#ifdef DEBUG
    printf("[Serveur] [Gestion_Client.c] [gestionClient] Sortie du client '%s'\n", s->pseudo);
#endif
    pthread_exit(0);
}

/* Permet de vérifier si le message du client n'est pas une requête serveur afin de l'envoyer vers la fonction de traitement */
int gestion_requete(char *message, info_client* client, char* protocole){
    char *chaine = NULL;
    
#ifdef DEBUG
    printf("[Serveur] [Gestion_Client.c] [gestion_requete] Message recu : '%s'\n",message);
#endif
    
    chaine = strstr(message,"Init_Pseudo!: "); //vérifie que la trame reçue est celle utilisée pour login un client
    if(chaine != NULL){
           
#ifdef DEBUG
    printf("[Serveur] [Gestion_Client.c] [gestion_requete] Détection commande login avec le message : '%s'\n",message);
#endif
    
        int i = login(message, client);
        return i;
    }
    int protection = protection_serveur(message, client);
    if(protection == 0){
        
#ifdef DEBUG
    printf("[Serveur] [Gestion_Client.c] [gestion_requete] Détection message erroné avec le message : '%s'\n",message);
#endif
    
        return -2;
    }
    chaine = strstr(message,"Co!: "); //vérifie que la trame reçue est celle utilisée pour ce connecter au client B
    if(chaine != NULL){
        
#ifdef DEBUG
    printf("[Serveur] [Gestion_Client.c] [gestion_requete] Détection commande de connexion à un autre client avec le message : '%s'\n",message);
#endif
    
        int i = Connexion(message, client, protocole); 
        return i;
    }
    chaine = strstr(message,"Deco!!!"); //vérifie que la trame reçue est celle utilisée pour la déconnexion
    if(chaine != NULL){
        
#ifdef DEBUG
    printf("[Serveur] [Gestion_Client.c] [gestion_requete] Détection commande de deconnexion avec le message : '%s'\n",message);
#endif
    
        return -1;
    }
    return 0;
}

/*Fonction permettant la protection du serveur évitant tous les crashs serveur connus*/
int protection_serveur(char* message, info_client* client){
    char* clientA = malloc(sizeof(char)*TAILLE_PSEUDO);
    int i = 0;
    
    while(message[i] != ':' && message[i] != '\0' && message[i] != '\n' && message[i] >=48 && message[i] <= 122){   //Récupération du login du client A 
        clientA[i] = message[i];
        i++;
    }
    clientA[i] = '\0';
    strcpy(client->pseudo,clientA);
    int j = hachage(*client);
    if(strcmp(liste_client[j].pseudo, clientA) != 0){
        
#ifdef DEBUG
    printf("[Serveur] [Gestion_Client.c] [protection_serveur] Pseudo associé au message introuvable dans la base de données, message reçu : '%s' pseudo détecté : '%s'\n",message, clientA);
#endif
    free(clientA);
        return 0;
    }
    
#ifdef DEBUG
    printf("[Serveur] [Gestion_Client.c] [protection_serveur] Pseudo reconnu, message avant traitement : '%s'\n",message);
#endif
    message = message + i + 1;
    
#ifdef DEBUG
    printf("[Serveur] [Gestion_Client.c] [protection_serveur] Message après traitement : '%s'\n",message);
#endif
    
    free(clientA);
    return 1;
}

/*Fonction de traitement lors de l'arrivée d'un nouveau client et de son login */
int login(char* message, info_client* client){
    char *chaine = NULL;
    chaine = strchr(message,' ');
    char* login = malloc(sizeof(char)*TAILLE_PSEUDO);
    int i = 1;
    while(chaine[i] != '\n'){   //Récupération du login que le client a envoyé
        login[i-1] = chaine[i];
        i++;
    }
    login[i-1] = '\0';
    
    printf("[Serveur] [Gestion_Client.c] [login] Connexion de '%s'\n",login);

    /* Sauvegarde de toute les informations nécessaires pour pouvoir initier une connexion avec celui ci */
    strcpy(client->pseudo,login);
    free(login);    
    if(Connexion_client(*client) != 0){
        printf("[Serveur] [Gestion_Client.c] [login] Le serveur est plein\n");
        return -1;
    }
    return 1;
}

int Connexion(char* message,info_client* client,char* protocole){
    char *chaine = NULL;
    int i = 0;
    chaine = strchr(message,' ');
    char* clientB = malloc(sizeof(char)*TAILLE_PSEUDO);
    char* clientA = malloc(sizeof(char)*TAILLE_PSEUDO);
    while(message[i] != ':'){   //Récupération du login du client A 
        clientA[i] = message[i];
        i++;
    }
    i = 1;
    while(chaine[i] != '\n'){   //Récupération du login que le client A a envoyé
        clientB[i-1] = chaine[i];
        i++;
    }
    clientB[i-1] = '\0';
    
    printf("[Serveur] [Gestion_Client.c] [Connexion] Connexion de '%s' à '%s'\n",clientA,clientB);
    /* Sauvegarde de toute les informations nécessaires pour pouvoir initier une connexion avec celui ci */

    Initiation_hole_punching(clientA,clientB,protocole);
    free(clientB);
    free(clientA);
    
    return 1;
}

/*Récupère les informations des pairs à partir de leur pseudo pour créer le message hole punching et leur envoyer*/
void Initiation_hole_punching(char* client_A, char* client_B, char* protocole){

#ifdef DEBUG
    printf("[Serveur] [Gestion_Client.c] [Initiation_hole_punching] Initiation du hole punching entre les clients '%s' et '%s' en '%s'\n",client_A,client_B,protocole);
#endif
    
    struct sockaddr_in ClientA;
    struct sockaddr_in ClientB;
    socklen_t lenA = sizeof(ClientA);
    socklen_t lenB = sizeof(ClientB);
    
    info_client Cb;
    info_client Ca;
    strcpy(Ca.pseudo,client_A);
    strcpy(Cb.pseudo,client_B);
    int a = hachage(Ca);
    int b = hachage(Cb);
    
    ClientA.sin_family = AF_INET;
    ClientA.sin_port = htons(liste_client[a].port);
    inet_aton(liste_client[a].hote, &ClientA.sin_addr);
    ClientB.sin_family = AF_INET;
    ClientB.sin_port = htons(liste_client[b].port);
    inet_aton(liste_client[b].hote, &ClientB.sin_addr);
    char* message = (char*) malloc(sizeof(char)*MAX_LIGNE);
    int n = sprintf(message,"Connexion_a!: %s\nIP!: %s\nport!: %d\n",client_A,liste_client[a].hote,liste_client[a].port);
    message[n] = '\0';
    int statut = sendto(liste_client[b].socket,message,MAX_LIGNE, 0, (struct sockaddr *) &ClientB, lenB);
    if(statut < 0){
        printf("socket : %d\n",liste_client[b].socket);
        perror("Initiation_hole_punching.sendto");
    }
    
#ifdef DEBUG
        printf("[Serveur] [Gestion_Client.c] [Initiation_hole_punching] Message envoyé à '%s' : '%s'\n",client_B, message);
#endif
        
    free(message);
    char* message_2 = (char*) malloc(sizeof(char)*MAX_LIGNE);
    n = sprintf(message_2,"Connexion_a!: %s\nIP!: %s\nport!: %d\n",client_B,liste_client[b].hote,liste_client[b].port);
    message_2[n] = '\0';
    statut = sendto(liste_client[a].socket,message_2,MAX_LIGNE, 0, (struct sockaddr *) &ClientA , lenA);
    if(statut < 0){
        printf("Socket : %d\n",liste_client[a].socket);
        perror("Initiation_hole_punching.sendto");
    }
    
#ifdef DEBUG
        printf("[Serveur] [Gestion_Client.c] [Initiation_hole_punching] Message envoyé à '%s' : '%s'\n",client_A, message_2);
#endif
        

}


/*Permet la déconnexion de tous les clients actuellement connectés au serveur afin de faire un arrêt propre */
void force_deconnexion_client_TCP(void){
    if( nombre_de_client_connecte() > 0){
        for(int i = 0;i < MAX_CLIENT_CONNECTE ; i++){
            if(liste_client[i].pseudo[0] != 0){
                char message[MAX_LIGNE] = "DeconnexionServeur!!!\n"; 
                
#ifdef DEBUG
                printf("[Serveur] [Gestion_Client.c] [force_deconnexion_client_TCP] Force la déconnexion de : '%s' avec le message : '%s'\n",liste_client[i].pseudo, message);
#endif
        
                write(liste_client[i].socket, message, MAX_LIGNE);
            }
        }
    }
}

/*Permet la déconnexion de tous les clients actuellement connectés au serveur afin de faire un arret propre */
void force_deconnexion_client_UDP(void){
    if( nombre_de_client_connecte() > 0){
        for(int i = 0;i < MAX_CLIENT_CONNECTE ; i++){
            if(liste_client[i].pseudo[0] != 0){
                
                struct sockaddr_in Client;
                socklen_t len = sizeof(Client);
                Client.sin_family = AF_INET;
                Client.sin_port = htons(liste_client[i].port);
                inet_aton(liste_client[i].hote, &Client.sin_addr);
                char message[MAX_LIGNE] = "DeconnexionServeur!!!\n";
                            
#ifdef DEBUG
                printf("[Serveur] [Gestion_Client.c] [force_deconnexion_client_UDP] Force la déconnexion de : '%s' qui a pour IP: '%s' et port : '%d' avec le message : '%s'\n",liste_client[i].pseudo , inet_ntoa(Client.sin_addr) , ntohs(Client.sin_port) , message);
#endif
        
                sendto(liste_client[i].socket, message, MAX_LIGNE,0,(struct sockaddr *)&Client,len);
            }
        }
    }
}
